/*
 * gsampler-midi-map.c
 * Copyright (C) Lilian Monchalin 2011 <isaya07@users.sf.net>
 *
 * gsampler is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * gsampler is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <math.h>
#include <stdlib.h>
#include <glib/gi18n-lib.h>

#include "gsampler-midi-map.h"
#include "gsampler-window.h"
#include "gsampler-midi-inst-dialog.h"
#include "gsampler-client-midimap.h"
#include "gsampler-client-channel.h"
#include "gsampler-instrument.h"
#include "gsampler-utils.h"

G_DEFINE_TYPE (GsamplerMidiMap, gsampler_midi_map, GTK_TYPE_BOX);

static GsamplerMidiMap* gsampler_midi_map__instance = NULL;
static gboolean gsampler_midi_map_status = FALSE;
static GSList* gsampler_midi_map_list = NULL;
static GSList* map_path_list = NULL;
static GHashTable* inst_path_list = NULL;

static void 
gsampler_midi_map_on_add_map_clicked (GtkButton* button, gpointer data) 
{
	GtkDialog* dialog;
	GsamplerMainWindow* window;
	GsamplerClient* client;
	GtkEntry* entry;
	GtkLabel* label;
	GtkBox* vbox1;
	gint result;
	GsamplerMidiMap* self = GSAMPLER_MIDI_MAP (data);
	g_return_if_fail (GSAMPLER_IS_MIDI_MAP (self));
	g_return_if_fail (button != NULL);
	
	window = gsampler_main_window_get_instance ();
	dialog = (GtkDialog*) gtk_dialog_new_with_buttons (_("Create Midi Instrument Map"),
	                                                   (GtkWindow*) window,
	                                                   GTK_DIALOG_DESTROY_WITH_PARENT, 
	                                                   GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, 
	                                                   GTK_STOCK_APPLY, GTK_RESPONSE_APPLY, NULL);
	
	entry = (GtkEntry*) gtk_entry_new ();
	label = (GtkLabel*) gtk_label_new ("Choose the name of the map");
	vbox1 = (GtkBox*) gtk_dialog_get_content_area (dialog);
	gtk_box_pack_start (vbox1, GTK_WIDGET (label), FALSE, TRUE, 5);
	gtk_box_pack_start (vbox1, GTK_WIDGET (entry), FALSE, TRUE, 5);
	gtk_entry_set_activates_default (entry, TRUE);
	gtk_dialog_set_default_response (dialog, GTK_RESPONSE_APPLY);
	gtk_widget_show_all ((GtkWidget*) dialog);

	gchar* name;
	gint map_id;
	result = gtk_dialog_run (dialog);
	switch (result) 
	{
		case GTK_RESPONSE_APPLY:
			name = g_strdup (gtk_entry_get_text (entry));
			client = gsampler_client_get_instance ();
			map_id = gsampler_client_create_midi_instrument_map (client, name);
			if (map_id == (-1)) 
			{
				gsampler_main_window_send_message ("warning", "Can't create the midi map ");
			}
			g_object_unref (client);
			g_free (name);
			break;
		case GTK_RESPONSE_CANCEL:
			break;
		default:
			break;
	}
	gtk_widget_destroy ((GtkWidget*) dialog);
	g_object_unref (window);
}

static void
gsampler_midi_map_on_edit_map_clicked (GtkButton* button, gpointer data)
{
	GtkTreeModel* model;
	GtkTreeIter iter = {0};
	GsamplerClient* client= NULL;
	GtkTreeSelection* select= NULL;
	GtkDialog *dialog = NULL;
	GtkWidget *ren_lbl = NULL;
	GtkEntry *ren_entry = NULL;
	GsamplerMainWindow* win = NULL;
	GtkBox* vbox1 = NULL;
	GsamplerMidiMap* self = GSAMPLER_MIDI_MAP (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (button != NULL);

	select = gtk_tree_view_get_selection (self->priv->map_tree);
	if (gtk_tree_selection_get_selected (select, &model, &iter)) 
	{
		gint id;
		gchar* name = NULL;
		gtk_tree_model_get (model, &iter, 0, &id, 1, &name, -1);

		if (id != -1 && name != NULL) {
			ren_lbl = gtk_label_new (_("Rename the midi map"));
			ren_entry = (GtkEntry*) gtk_entry_new ();
			gtk_entry_set_text (ren_entry, name);
			client = gsampler_client_get_instance ();
			win = gsampler_main_window_get_instance ();
			dialog = (GtkDialog*) gtk_dialog_new_with_buttons (_("Rename"), (GtkWindow*) win, 
	                                      			   		   GTK_DIALOG_DESTROY_WITH_PARENT, 
	                                      			           GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
			                                                   GTK_STOCK_APPLY, GTK_RESPONSE_APPLY,
	                                      			           NULL);
			vbox1 = (GtkBox*) gtk_dialog_get_content_area (dialog);
			gtk_box_pack_start (vbox1, ren_lbl, FALSE, TRUE, 0);
			gtk_box_pack_start (vbox1, GTK_WIDGET (ren_entry), FALSE, TRUE, 0);
			gtk_widget_show_all ((GtkWidget*) dialog);
			gint result = gtk_dialog_run (dialog);
			if (result == GTK_RESPONSE_APPLY) 
			{
				gchar* result = NULL;
				result = g_strdup (gtk_entry_get_text (ren_entry));
				if (g_strcmp0 (name, result) != 0)
				{
					if (!gsampler_client_set_midi_instrument_map_name (client, id, result));
						g_warning ("Can't rename midi map");
				}
				g_free (result);
			}
			gtk_widget_destroy ((GtkWidget*) dialog);
			g_object_unref (win);
			g_object_unref (client);
			g_free (name);
		}
	}
}

static void 
gsampler_midi_map_on_del_map_clicked (GtkButton* button, gpointer data) 
{
	GtkTreeModel* model;
	GtkTreeIter iter = {0};
	GsamplerClient* client;
	GtkTreeSelection* select;
	GsamplerMidiMap* self = GSAMPLER_MIDI_MAP (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (button != NULL);
	
	client = gsampler_client_get_instance ();
	select = gtk_tree_view_get_selection (self->priv->map_tree);
	if (gtk_tree_selection_get_selected (select, &model, &iter)) 
	{
		gint id;
		gtk_tree_model_get (model, &iter, 0, &id, -1);
		if (!gsampler_client_delete_midi_instrument_map (client, id)) {
			gsampler_main_window_send_message ("warning", "Can't delete Midi Instrument Map");
		}
	}
	g_object_unref (client);
}

static void 
gsampler_midi_map_update_midi_instrument (GsamplerMidiMap* self, GSList* new_list) {
	GtkTreeIter iter = {0};
	GtkTreeIter parent = {0};
	GsamplerClient* client;
	GtkTreeStore* store;
	g_return_if_fail (self != NULL);
	
	client = gsampler_client_get_instance ();
	store = (GtkTreeStore*) gtk_tree_view_get_model (self->priv->inst_tree);
	gtk_tree_store_clear (store);
	g_hash_table_remove_all (inst_path_list);

	while (new_list) 
	{
		gchar* inst = NULL;
		inst = g_strdup (new_list->data);
		if (inst != NULL && g_strcmp0 (inst, "") != 0) 
		{
			gchar** tmp;
			gint bank_id, inst_id;
			gchar* inst_name;
			GtkTreePath* path_bank = NULL;
			GsamplerMidiInstInfo* info;

			tmp = g_strsplit (inst, ",", 0);
			bank_id = atoi (tmp[1]);
			inst_id = atoi (tmp[2]);
			path_bank =  gtk_tree_path_new_from_indices (bank_id, -1);
			gtk_tree_model_get_iter ((GtkTreeModel*) store, &parent, path_bank);
			if (!gtk_tree_store_iter_is_valid (store, &parent)) {
				gchar* bank_name;
				gtk_tree_store_append (store, &parent, NULL);
				bank_name = g_strconcat ("bank ", tmp[1], NULL);
				gtk_tree_store_set (store, &parent, 0, bank_id, 1, bank_name, -1, -1);
				g_free (bank_name);
			}
			info = gsampler_client_get_midi_instruments_info (client, atoi (tmp[0]), bank_id, inst_id);
			gtk_tree_store_append (store, &iter, &parent);
			GtkTreePath* path_inst = NULL;
			path_inst = gtk_tree_model_get_path ((GtkTreeModel*) store, &iter);
			inst_name = g_strconcat (tmp[2], " ", info->name, NULL);
			gtk_tree_store_set (store, &iter, 0, inst_id, 1, inst_name, -1, -1);
			g_hash_table_insert (inst_path_list, g_strdup_printf ("%d,%d", bank_id, inst_id),
			                     gtk_tree_path_copy (path_inst));
			gtk_tree_path_free (path_inst);
			g_free (inst_name);
			gsampler_midi_inst_info_free (info);
			gtk_tree_path_free (path_bank);
			g_strfreev (tmp);
		}
		g_free (inst);
		new_list = new_list->next;
	}
	gtk_tree_view_expand_all (self->priv->inst_tree);
	g_object_unref (client);
}

static void 
gsampler_midi_map_on_add_instrument_clicked (GtkButton* button, gpointer data) 
{
	GsamplerMidiInstDialog* dialog = NULL;
	GtkTreeIter iter = {0};
	GtkTreeModel* model;
	gint map = -1;
	GsamplerClient* client;
	GsamplerMainWindow* window;
	gchar** engine_list;
	gint engine_list_lenght = 0;
	GtkTreeSelection* select;
	gint result;
	GsamplerMidiMap* self = GSAMPLER_MIDI_MAP (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (button != NULL);
	
	client = gsampler_client_get_instance ();
	window = gsampler_main_window_get_instance ();
	dialog = gsampler_midi_inst_dialog_new (FALSE);
	gtk_window_set_transient_for ((GtkWindow*) dialog, (GtkWindow*) window);
	engine_list = gsampler_client_get_available_engine_list (client);
	
	if (engine_list != NULL)
		engine_list_lenght = g_strv_length (engine_list);
	gint element_it;
	for (element_it = 0; element_it < engine_list_lenght; element_it = element_it + 1) 
	{
		gchar* element;
		element = g_strdup (engine_list[element_it]);
		gsampler_text_combo_append (dialog->engine, string_replace (element, "'", ""));
		g_free (element);
	}
	gtk_combo_box_set_active ((GtkComboBox*) dialog->engine, 0);
	select = gtk_tree_view_get_selection (self->priv->map_tree);
	if (gtk_tree_selection_get_selected (select, &model, &iter)) 
	{
		gtk_tree_model_get (model, &iter, 0, &map, -1);
		if (map == (-1)) {
			gsampler_main_window_send_message ("warning", _("Invalide MidiMap"));
			return;
		} else {
			result = gtk_dialog_run ((GtkDialog*) dialog);
			if (result == GTK_RESPONSE_APPLY) {
				gchar* file;
				gint instrument;
				gchar* engine;
				gchar* load_mode;
				gchar* name;
				gdouble vol;
				gdouble bank;
				gint prog;

#ifdef G_OS_WIN32
				file = string_replace (gtk_file_chooser_get_filename ((GtkFileChooser*) dialog->file), "\\", "/");
#else
				file = gtk_file_chooser_get_filename ((GtkFileChooser*) dialog->file);
#endif
				instrument = gtk_combo_box_get_active ((GtkComboBox*) dialog->inst);
				engine = gsampler_text_combo_get_active (dialog->engine);
				load_mode = gsampler_text_combo_get_active (dialog->load_combo);
				name = g_strdup (gtk_entry_get_text (dialog->name_entry));
				vol = gtk_adjustment_get_value (dialog->vol_adj);
				bank = gtk_spin_button_get_value (dialog->bank_spin);
				prog = atoi (gsampler_text_combo_get_active (dialog->prog_combo));
				if (gsampler_client_map_midi_instrument (client, map, bank, prog, engine, file,
						                                  instrument, vol, load_mode, name)) {
					/*GSList *inst_list = NULL;
					inst_list = gsampler_client_get_available_midi_instruments (client, map);
					gsampler_midi_map_update_midi_instrument (self, inst_list);*/
					gtk_tree_selection_unselect_all (select);
					gtk_tree_selection_select_iter (select, &iter);										  
				} else {
					gsampler_main_window_send_message ("warning", _("Can't Map Midi Instrument"));
				}
				g_free (name);
				g_free (load_mode);
				g_free (engine);
				g_free (file);
			}
		}
	}
	gtk_widget_destroy ((GtkWidget*) dialog);
	//g_object_unref (select);
	g_strfreev (engine_list);
	g_object_unref (window);
	g_object_unref (client);
}

static gboolean 
gsampler_midi_map_insert_new_map (GsamplerMidiMap* self, gint map_id) 
{
	GsamplerClient* client;
	GsamplerMidiMapInfo *info = NULL;
	GtkTreeIter iter = {0};
	GtkListStore* list;
	GtkTreeSelection *select;
	g_return_val_if_fail (self != NULL, FALSE);

	if (g_id_list_exist (gsampler_midi_map_list, map_id)) 
	{
		return TRUE;
	}
	client = gsampler_client_get_instance ();
	info = gsampler_client_get_midi_instrument_map_info (client, map_id);
	list = (GtkListStore*) gtk_tree_view_get_model (self->priv->map_tree);
	gtk_list_store_insert (list, &iter, map_id);
	gtk_list_store_set (list, &iter, 0, map_id, 1, g_strdup (info->name), -1);
	GtkTreePath *path = gtk_tree_model_get_path ((GtkTreeModel*) list, &iter);
	gsampler_midi_map_list = g_id_list_add (gsampler_midi_map_list, map_id, info->name);
	map_path_list = g_id_list_add (map_path_list, map_id, path);
	select = gtk_tree_view_get_selection (self->priv->map_tree);
	gtk_tree_selection_select_iter (select, &iter);
		
	gsampler_midi_map_info_free (info);
	g_object_unref (client);
	return TRUE;
}

static void 
gsampler_midi_map_on_map_changed (GtkTreeSelection* select, gpointer data) 
{
	GtkTreeIter iter = {0};
	GtkTreeModel* model = NULL;
	GsamplerMidiMap* self = GSAMPLER_MIDI_MAP (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (select != NULL);

	if (gtk_tree_selection_get_selected (select, &model, &iter)) 
	{
		gint id;
		gchar* name;
		GsamplerClient* client;
		GSList *list = NULL;
		
		client = gsampler_client_get_instance ();
		gtk_tree_model_get (model, &iter, 0, &id, 1, &name, -1);
		list = gsampler_client_get_available_midi_instruments (client, id);
		gsampler_midi_map_update_midi_instrument (self, list);
		gtk_widget_set_sensitive ((GtkWidget*) self->priv->inst_button, TRUE);
		gtk_widget_set_sensitive ((GtkWidget*) self->priv->inst_tree, TRUE);
		gtk_widget_set_sensitive ((GtkWidget*) self->priv->del_map_button, TRUE);
		gtk_widget_set_sensitive ((GtkWidget*) self->priv->edit_map_button, TRUE);
		g_slist_free_full (list, g_free);
		g_object_unref (client);
		g_free (name);
	} else
	{
		gsampler_midi_map_update_midi_instrument (self, NULL);
		gtk_widget_set_sensitive ((GtkWidget*) self->priv->inst_button, FALSE);
		gtk_widget_set_sensitive ((GtkWidget*) self->priv->inst_tree, FALSE);
		gtk_widget_set_sensitive ((GtkWidget*) self->priv->del_map_button, FALSE);
		gtk_widget_set_sensitive ((GtkWidget*) self->priv->edit_map_button, FALSE);
	}
}

static void 
gsampler_midi_map_on_midi_instrument_count (GsamplerClient* client, 
                                            GSList* new_list, gpointer data) {
	GtkTreeSelection* select;
	GsamplerMidiMap* self = GSAMPLER_MIDI_MAP (data);
	g_return_if_fail (self != NULL);
	
	while (!gsampler_midi_map_status) 
	{
		g_usleep ((guint) 300000);
	}
	gsampler_midi_map_status = FALSE;
	select = gtk_tree_view_get_selection (self->priv->map_tree);
	if (gtk_tree_selection_get_selected (select, NULL, NULL)) 
	{
		gsampler_midi_map_update_midi_instrument (self, new_list);
	}
	gsampler_midi_map_status = TRUE;
}

static void
gsampler_midi_map_on_midi_instrument_info (GsamplerClient* client, 
                                           gint map_id, gint bank, gint prog, gpointer data)
{
	GtkTreeIter iter = {0};
	GtkTreePath *path = NULL;
	GtkTreeStore *store = NULL;
	GsamplerMidiMap* self = GSAMPLER_MIDI_MAP (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (client != NULL);

	path = g_hash_table_lookup (inst_path_list, g_strdup_printf ("%d,%d", bank, prog));
	if (path != NULL)
	{
		store = (GtkTreeStore*) gtk_tree_view_get_model (self->priv->inst_tree);
		if (gtk_tree_model_get_iter ((GtkTreeModel*) store, &iter, path))
		{
			GsamplerMidiInstInfo *info = NULL;
			gchar* inst_name = NULL;

			info = gsampler_client_get_midi_instruments_info (client, map_id, bank, prog);
			inst_name = g_strdup_printf ("%d %s", prog, info->name);
			gtk_tree_store_set (store, &iter, 0, prog, 1, inst_name,-1);
			gsampler_midi_inst_info_free (info);
			g_free (inst_name);
		} else 
		{
			g_warning ("invalide iter for update midi instrument");
		}
	} else 
	{
		g_warning ("invalide path for update midi instrument");
	}
}

static void 
gsampler_midi_map_on_map_count (GsamplerClient* client, gchar** new_list, gpointer data) 
{
	gint new_list_length = 0;
	GsamplerMidiMap* self = GSAMPLER_MIDI_MAP (data);
	g_return_if_fail (self != NULL);
	
	while (!gsampler_midi_map_status) 
	{
		g_usleep ((guint) 300000);
	}
	gsampler_midi_map_status = FALSE;

	if (new_list != NULL)
		new_list_length = g_strv_length (new_list);

	GtkListStore* store;
	store = (GtkListStore*) gtk_tree_view_get_model (self->priv->map_tree);
	gtk_list_store_clear (store);
	g_id_list_free (gsampler_midi_map_list, TRUE);
	gsampler_midi_map_list = NULL;

	int map_it;
	for (map_it = 0; map_it < new_list_length; map_it = map_it + 1) 
	{
		gchar* map;
		map = g_strdup (new_list[map_it]);
		if (g_strcmp0 (map, "NONE") != 0) 
		{
			gint map_id;
			map_id = atoi (map);
			if (!g_id_list_exist (gsampler_midi_map_list, map_id)) 
			{
				if (!gsampler_midi_map_insert_new_map (self, map_id)) 
					g_warning ("Can't insert midimap");
			}
		}
		g_free (map);	
	}
	gsampler_midi_map_status = TRUE;
}

static void 
gsampler_midi_map_on_map_info (GsamplerClient* client, gint id, gpointer data)
{
	GtkListStore *list = NULL;
	GtkTreeIter iter = {0};
	GtkTreePath *path = NULL;
	GsamplerMidiMap* self = GSAMPLER_MIDI_MAP (data);
	g_return_if_fail (self != NULL);

	
	path = (GtkTreePath *) g_id_list_get (map_path_list, id);
	if (path != NULL)
	{
		list = (GtkListStore*) gtk_tree_view_get_model (self->priv->map_tree);
		if (gtk_tree_model_get_iter ((GtkTreeModel*) list, &iter, path))
		{
			GsamplerMidiMapInfo *info = NULL;
			gchar* map_name = NULL;

			info = gsampler_client_get_midi_instrument_map_info (client, id);
			map_name = g_strdup (info->name);
			gtk_list_store_set (list, &iter, 0, id, 1, map_name,-1);
			gsampler_midi_map_info_free (info);
			g_free (map_name);
		} else 
		{
			g_warning ("invalide iter for update midi map.");
		}
	} else 
	{
		g_warning ("invalide path for update midi map.");
	}
}


static void 
gsampler_midi_map_edit_midi_instrument (GtkMenuItem* item, gpointer data) 
{
	GsamplerMidiInstDialog* dialog = NULL;
	GtkTreeIter iter = {0};
	GtkTreeIter parent = {0};
	GtkTreeModel* model = NULL;
	gint bank_id;
	gint prog_id;
	gint map_id = -1;
	gchar* inst = NULL;
	GsamplerClient* client = NULL;
	GsamplerMainWindow* window = NULL;
	GtkTreeSelection* select;
	GsamplerMidiInstInfo* info;
	gchar** engine_list = NULL;
	gint engine_list_lenght = 0;
	gint result;
	GsamplerMidiMap* self = GSAMPLER_MIDI_MAP (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (item != NULL);

	client = gsampler_client_get_instance ();
	window = gsampler_main_window_get_instance ();
	dialog = gsampler_midi_inst_dialog_new (TRUE);
	gtk_window_set_transient_for ((GtkWindow*) dialog, (GtkWindow*) window);
	select = gtk_tree_view_get_selection (self->priv->map_tree);
	if (gtk_tree_selection_get_selected (select, &model, &iter)) 
		gtk_tree_model_get (model, &iter, 0, &map_id, -1);
	
	if (map_id == (-1)) 
	{
		gsampler_main_window_send_message ("warning",  _("Invalide MidiMap"));
		g_object_unref (window);
		g_object_unref (client);
		g_object_unref (dialog);
		return;
	}
	select = gtk_tree_view_get_selection (self->priv->inst_tree);
	if (gtk_tree_selection_get_selected (select, &model, &iter)) 
	{
		GtkTreeStore* store;
		store = (GtkTreeStore*) gtk_tree_view_get_model (self->priv->inst_tree);
		gtk_tree_model_get ((GtkTreeModel*) store, &iter, 0, &prog_id, 1, &inst, -1);
		gtk_tree_model_iter_parent ((GtkTreeModel*) store, &parent, &iter);
		if (gtk_tree_store_iter_is_valid (store, &parent)) 
		{
			gtk_tree_model_get ((GtkTreeModel*) store, &parent, 0, &bank_id, -1);
		} else 
		{
			gsampler_main_window_send_message ("warning", g_strconcat ("Can't supress ", inst, NULL));
			g_object_unref (window);
			g_object_unref (client);
			g_free (inst);
			g_object_unref (dialog);
			return;
		}
	} else 
	{
		gsampler_main_window_send_message ("warning", _("Invalide Midi Instrument"));
		g_object_unref (window);
		g_object_unref (client);
		g_free (inst);
		g_object_unref (dialog);
		return;
	}
	info = gsampler_client_get_midi_instruments_info (client, map_id, bank_id, prog_id);
	if (g_strcmp0 (info->name, "NONE") == 0) 
	{
		gsampler_main_window_send_message ("warning", 
		                                   g_strconcat ("Can' get info for inst ", inst, NULL));
		gsampler_midi_inst_info_free (info);
		g_object_unref (window);
		g_object_unref (client);
		g_free (inst);
		g_object_unref (dialog);
		return;
	}
	
	engine_list = gsampler_client_get_available_engine_list (client);
	if (engine_list != NULL)
		engine_list_lenght = g_strv_length (engine_list);
	gint i = 0;
	gint element_it;
	for (element_it = 0; element_it < engine_list_lenght; element_it = element_it + 1) 
	{
		gchar* element;
		element = string_replace (engine_list[element_it], "'", "");
		gsampler_text_combo_append (dialog->engine, element);
		if (g_strcmp0 (element, info->engine) == 0)
			gtk_combo_box_set_active ((GtkComboBox*) dialog->engine, i);
		i++;
		g_free (element);
	}

	GsamplerInstrument* instrument;
	instrument = gsampler_instrument_new ();
	if (gsampler_instrument_load_file (instrument, info->file)) 
	{
		GSList* inst_list = NULL;

		gsampler_text_combo_remove_all (dialog->inst);
		inst_list = gsampler_instrument_get_list (instrument);
		while (inst_list) 
		{
			struct Struct_list_widget* instru;
			instru = (struct Struct_list_widget*) inst_list->data;
			gsampler_text_combo_insert (dialog->inst, (const gchar*) instru->Sdata, instru->Sid);
			inst_list = inst_list->next;
		}
	}
	
	gtk_file_chooser_set_filename ((GtkFileChooser*) dialog->file, info->file);
	gtk_adjustment_set_value (dialog->vol_adj, (info->vol*100));
	gtk_combo_box_set_active ((GtkComboBox*) dialog->inst, info->inst_nb);
	gtk_entry_set_text (dialog->name_entry, info->name);
	gsampler_text_combo_set_active (dialog->engine, info->engine);
	gsampler_text_combo_set_active (dialog->load_combo, info->load_mode);
	
	gsampler_midi_inst_dialog_connect (dialog);
	result = gtk_dialog_run ((GtkDialog*) dialog);
	if (result == GTK_RESPONSE_APPLY) {
		gchar* file = gtk_file_chooser_get_filename ((GtkFileChooser*) dialog->file);
		gint instrument = gtk_combo_box_get_active ((GtkComboBox*) dialog->inst);
		gchar* engine = gsampler_text_combo_get_active (dialog->engine);
		gchar* load_mode = gsampler_text_combo_get_active (dialog->load_combo);
		gchar* name = g_strdup (gtk_entry_get_text (dialog->name_entry));
		gdouble vol = gtk_adjustment_get_value (dialog->vol_adj);

		if (!gsampler_client_map_midi_instrument (client, map_id, bank_id, prog_id, engine, file,
						                                  instrument, vol, load_mode, name))
		{
			gsampler_main_window_send_message ("warning", _("Can't Map Midi Instrument"));
		}		
	}
	gtk_widget_destroy ((GtkWidget*) dialog);
	g_strfreev (engine_list);
	gsampler_midi_inst_info_free (info);
	g_object_unref (window);
	g_object_unref (client);
	g_free (inst);
}

static void 
gsampler_midi_map_delete_midi_instrument (GtkMenuItem* item, gpointer data) {
	GtkTreeIter iter = {0};
	GtkTreeIter parent = {0};
	GtkTreeModel* model = NULL;
	gint bank_id = 0;
	gint prog_id = 0;
	gint map_id = 0;
	gchar* inst = NULL;
	GsamplerClient* client;
	GtkTreeSelection* select;
	GsamplerMidiMap* self = GSAMPLER_MIDI_MAP (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (item != NULL);
	
	client = gsampler_client_get_instance ();
	select = gtk_tree_view_get_selection (self->priv->inst_tree);
	if (gtk_tree_selection_get_selected (select, &model, &iter)) {
		GtkTreeStore* store;
		store = (GtkTreeStore*) gtk_tree_view_get_model (self->priv->inst_tree);
		gtk_tree_model_get ((GtkTreeModel*) store, &iter, 0, &prog_id, 1, &inst, -1);
		gtk_tree_model_iter_parent ((GtkTreeModel*) store, &parent, &iter);
		if (gtk_tree_store_iter_is_valid (store, &parent)) {
			gtk_tree_model_get ((GtkTreeModel*) store, &parent, 0, &bank_id, -1);
		} else {
			gsampler_main_window_send_message ("warning", g_strconcat ("Can't supress ", inst, NULL));
			g_object_unref (client);
			g_free (inst);
			return;
		}
	}
	select = gtk_tree_view_get_selection (self->priv->map_tree);
	if (gtk_tree_selection_get_selected (select, &model, &iter)) {
		gtk_tree_model_get (model, &iter, 0, &map_id, -1);
	} else {
		gsampler_main_window_send_message ("warning", g_strconcat ("Can't supress ", inst, NULL));
		g_object_unref (client);
		g_free (inst);
		return;
	}
	if (map_id == (-1)) {
		gsampler_main_window_send_message ("warning", _("Invalide MidiMap"));
		g_object_unref (client);
		g_free (inst);
		return;
	}
	if (!gsampler_client_unmap_midi_instrument (client, map_id, bank_id, prog_id)) {
		gsampler_main_window_send_message ("warning", g_strconcat ("Can't supress ", inst, NULL));
	}
	g_object_unref (client);
	g_free (inst);
}

static void 
gsampler_midi_map_do_popup_inst (GsamplerMidiMap* self, GdkEventButton* event) {
	guint button = 0U;
	guint32 event_time = 0U;
	GtkTreeModel* model = NULL;
	GtkTreeIter iter = {0};
	GtkTreeSelection* select;
	GtkTreeStore* store;
	GtkMenu* new_menu;
	GtkMenuItem* new_item;
	g_return_if_fail (self != NULL);
	
	select = gtk_tree_view_get_selection (self->priv->inst_tree);
	if (!gtk_tree_selection_get_selected (select, &model, &self->priv->current_iter)) {
		return;
	}
	store = (GtkTreeStore*)  gtk_tree_view_get_model (self->priv->inst_tree);
	gtk_tree_model_iter_parent ((GtkTreeModel*) store, &iter, &self->priv->current_iter);
	if (! gtk_tree_store_iter_is_valid (store, &iter)) {
		return;
	}
	if (event != NULL) {
		button = event->button;
		event_time = event->time;
	} else {
		button = 0;
		event_time = gtk_get_current_event_time ();
	}
	new_menu = (GtkMenu*) gtk_menu_new ();
	new_item = (GtkMenuItem*) gtk_menu_item_new_with_label (_("Edit"));
	g_signal_connect_object (new_item, "activate", 
	                         (GCallback) gsampler_midi_map_edit_midi_instrument, self, 0);
	gtk_menu_shell_append ((GtkMenuShell*) new_menu, (GtkWidget*) new_item);
	new_item = (GtkMenuItem*) gtk_menu_item_new_with_label (_("Delete"));
	g_signal_connect_object (new_item, "activate", 
	                         (GCallback) gsampler_midi_map_delete_midi_instrument, self, 0);
	gtk_menu_shell_append ((GtkMenuShell*) new_menu, (GtkWidget*) new_item);
	g_object_set (new_menu, "attach-widget", (GtkWidget*) self->priv->inst_tree, NULL);
	gtk_widget_show_all ((GtkWidget*) new_menu);
	gtk_menu_popup (new_menu, NULL, NULL, NULL, NULL, button, event_time);
	gtk_tree_selection_select_iter (select, &self->priv->current_iter);
}

static gboolean 
gsampler_midi_map_on_inst_button_press_event (GtkWidget* sender, GdkEventButton* event, gpointer data) {
	GsamplerMidiMap* self = GSAMPLER_MIDI_MAP (data);
	g_return_val_if_fail (self != NULL, FALSE);
	
	if (event->type == GDK_BUTTON_PRESS && event->button == 3) {
		gsampler_midi_map_do_popup_inst (self, event);
		return TRUE;
	}
	return FALSE;
}

static gboolean 
gsampler_midi_map_on_popup_inst (GtkWidget* sender, gpointer data) {
	GsamplerMidiMap* self = GSAMPLER_MIDI_MAP (data);
	g_return_val_if_fail (self != NULL, FALSE);
	
	gsampler_midi_map_do_popup_inst (self, NULL);
	return TRUE;
}

static GObject *
gsampler_midi_map_constructor (GType type, guint n_construct_properties, 
                               GObjectConstructParam * construct_properties) 
{
	GObject * obj;
	GsamplerMidiMap * self = NULL;
	GtkWidget* hbox1 = NULL;
	GtkWidget* hbox2 = NULL;
	GtkLabel* map_lbl = NULL;
	GtkImage* addmap_image = NULL;
	GtkImage* edit_image = NULL;
	GtkImage* delmap_image = NULL;
	GtkLabel* inst_lbl = NULL;
	GtkListStore* liste = NULL;
	GtkTreeStore* store = NULL;
	GtkTreeSelection* select;

	obj = G_OBJECT_CLASS (gsampler_midi_map_parent_class)->constructor (type, n_construct_properties, construct_properties);
	self = GSAMPLER_MIDI_MAP (obj);
	gtk_orientable_set_orientation ((GtkOrientable*) self, GTK_ORIENTATION_VERTICAL);
	hbox1 = gtk_hbox_new (FALSE, 5);
	map_lbl = (GtkLabel*) gtk_label_new (NULL);
	gtk_label_set_markup (map_lbl, _("<b>Midi Maps :</b>"));
	addmap_image = (GtkImage*) gtk_image_new_from_stock (GTK_STOCK_ADD, GTK_ICON_SIZE_BUTTON);
	self->priv->add_map_button = (GtkButton*) gtk_button_new ();
	gtk_button_set_image (self->priv->add_map_button, (GtkWidget*) addmap_image);

	edit_image = (GtkImage*) gtk_image_new_from_stock (GTK_STOCK_EDIT, GTK_ICON_SIZE_BUTTON);
	self->priv->edit_map_button = (GtkButton*) gtk_button_new ();
	gtk_button_set_image (self->priv->edit_map_button, (GtkWidget*) edit_image);
	
	delmap_image = (GtkImage*) gtk_image_new_from_stock (GTK_STOCK_DELETE, GTK_ICON_SIZE_BUTTON);
	self->priv->del_map_button = (GtkButton*) gtk_button_new ();
	gtk_button_set_image (self->priv->del_map_button, (GtkWidget*) delmap_image);
	liste = gtk_list_store_new (2, G_TYPE_INT, G_TYPE_STRING);
	self->priv->map_tree = (GtkTreeView*) gtk_tree_view_new_with_model ((GtkTreeModel*) liste);
	g_object_unref(liste); //TODO
	gtk_tree_view_set_headers_visible (self->priv->map_tree, FALSE);
	/*gtk_tree_view_insert_column_with_attributes (self->priv->map_tree, -1, NULL,
	                                             (GtkCellRenderer*) gtk_cell_renderer_text_new (),
	                                             "text", 0, NULL, NULL);	*/
	gtk_tree_view_insert_column_with_attributes (self->priv->map_tree, -1, NULL,
	                                             (GtkCellRenderer*) gtk_cell_renderer_text_new (),
	                                             "text", 1, NULL, NULL);
	hbox2 = gtk_hbox_new (FALSE, 5);
	inst_lbl = (GtkLabel*) gtk_label_new (NULL);
	gtk_label_set_markup (inst_lbl, _("<b>Instruments :</b>"));
	self->priv->inst_button = (GtkButton*) gtk_button_new_from_stock (GTK_STOCK_ADD);
	store = gtk_tree_store_new (2, G_TYPE_INT, G_TYPE_STRING);
	self->priv->inst_tree = (GtkTreeView*) gtk_tree_view_new_with_model ((GtkTreeModel*) store);
	g_object_unref(store);
	gtk_tree_view_set_headers_visible (self->priv->inst_tree, FALSE);
	gtk_tree_view_set_hover_selection (self->priv->inst_tree, TRUE);
	gtk_tree_view_insert_column_with_attributes (self->priv->inst_tree, -1, NULL,
	                                             (GtkCellRenderer*) gtk_cell_renderer_text_new (),
	                                             "text", 1, NULL, NULL);
	select = gtk_tree_view_get_selection (self->priv->map_tree);
								   
	gtk_box_pack_start ((GtkBox*) hbox1, (GtkWidget*) map_lbl, TRUE, TRUE, 5);
	gtk_box_pack_start ((GtkBox*) hbox1, (GtkWidget*) self->priv->add_map_button, FALSE, TRUE, 5);
	gtk_box_pack_start ((GtkBox*) hbox1, (GtkWidget*) self->priv->edit_map_button, FALSE, TRUE, 5);
	gtk_box_pack_start ((GtkBox*) hbox1, (GtkWidget*) self->priv->del_map_button, FALSE, TRUE, 5);
	gtk_box_pack_start ((GtkBox*) hbox2, (GtkWidget*) inst_lbl, TRUE, TRUE, 5);
	gtk_box_pack_start ((GtkBox*) hbox2, (GtkWidget*) self->priv->inst_button, FALSE, TRUE, 5);
	gtk_box_pack_start ((GtkBox*) self, (GtkWidget*) hbox1, FALSE, TRUE, 0);
	gtk_box_pack_start ((GtkBox*) self, (GtkWidget*) self->priv->map_tree, TRUE, TRUE, 5);
	gtk_box_pack_start ((GtkBox*) self, (GtkWidget*) hbox2, FALSE, TRUE, 0);
	gtk_box_pack_start ((GtkBox*) self, (GtkWidget*) self->priv->inst_tree, TRUE, TRUE, 5);

	inst_path_list = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, 
	                                        (GDestroyNotify) gtk_tree_path_free);
								   
	g_signal_connect_object (select, "changed", 
	                         (GCallback) gsampler_midi_map_on_map_changed, self, 0);
	g_signal_connect_object (self->priv->add_map_button, "clicked", 
	                         (GCallback) gsampler_midi_map_on_add_map_clicked, self, 0);
	g_signal_connect_object (self->priv->edit_map_button, "clicked", 
	                         (GCallback) gsampler_midi_map_on_edit_map_clicked, self, 0);
	g_signal_connect_object (self->priv->del_map_button, "clicked", 
	                         (GCallback) gsampler_midi_map_on_del_map_clicked, self, 0);
	g_signal_connect_object (self->priv->inst_button, "clicked", 
	                         (GCallback) gsampler_midi_map_on_add_instrument_clicked, self, 0);
	g_signal_connect_object ((GtkWidget*) self->priv->inst_tree, "popup-menu", 
	                         (GCallback) gsampler_midi_map_on_popup_inst, self, 0);
	g_signal_connect_object ((GtkWidget*) self->priv->inst_tree, "button-press-event", 
	                         (GCallback) gsampler_midi_map_on_inst_button_press_event, self, 0);
	
	gtk_widget_set_sensitive ((GtkWidget*) self->priv->inst_tree, FALSE);
	gtk_widget_set_sensitive ((GtkWidget*) self->priv->inst_button, FALSE);
	gtk_widget_set_sensitive ((GtkWidget*) self->priv->del_map_button, FALSE);
	gtk_widget_set_sensitive ((GtkWidget*) self->priv->edit_map_button, FALSE);
	gtk_widget_set_sensitive ((GtkWidget*) self, FALSE);

	gtk_widget_show_all ((GtkWidget*) self);

	return obj;
}

static void 
gsampler_midi_map_init (GsamplerMidiMap * self) {
	self->priv = GSAMPLER_MIDI_MAP_GET_PRIVATE (self);
}

static void 
gsampler_midi_map_finalize (GObject* obj) {
	GsamplerMidiMap * self;
	self = GSAMPLER_MIDI_MAP (obj);
	g_object_unref (self->priv->map_tree);
	g_object_unref (self->priv->inst_tree);
	g_object_unref (self->priv->inst_button);
	g_object_unref (self->priv->del_map_button);
	g_object_unref (self->priv->edit_map_button);
	g_object_unref (self->priv->add_map_button);
	G_OBJECT_CLASS (gsampler_midi_map_parent_class)->finalize (obj);
}

static void 
gsampler_midi_map_class_init (GsamplerMidiMapClass * klass) {
	gsampler_midi_map_parent_class = g_type_class_peek_parent (klass);
	g_type_class_add_private (klass, sizeof (GsamplerMidiMapPrivate));
	G_OBJECT_CLASS (klass)->finalize = gsampler_midi_map_finalize;
	G_OBJECT_CLASS (klass)->constructor = gsampler_midi_map_constructor;
}

GsamplerMidiMap* 
gsampler_midi_map_new (void) {
	GsamplerMidiMap* self = NULL;
	
	self = (GsamplerMidiMap*) g_object_new (GSAMPLER_TYPE_MIDI_MAP, NULL);
	return self;
}

GsamplerMidiMap* 
gsampler_midi_map_get_instance (void) {
	if (gsampler_midi_map__instance == NULL) {
		gsampler_midi_map__instance = gsampler_midi_map_new ();
	}
	return g_object_ref (gsampler_midi_map__instance);
}

GSList * 
gsampler_midi_map_get_midi_maps (void) {
	// TODO update for session
	return gsampler_midi_map_list;
}

void 
gsampler_midi_map_active (GsamplerMidiMap* self) {
	GsamplerClient* client = NULL;
	gchar** map_list;
	g_return_if_fail (self != NULL);
	
	client = gsampler_client_get_instance ();
	gsampler_midi_map_status = TRUE;
	map_list = gsampler_client_get_available_midi_instrument_map (client);
	gsampler_midi_map_on_map_count (client, map_list, self);
	g_signal_connect_object (client, "map-count-event", 
	                         (GCallback) gsampler_midi_map_on_map_count, self, 0);
	g_signal_connect_object (client, "midi-instrument-count-event", 
	                         (GCallback) gsampler_midi_map_on_midi_instrument_count, self, 0);
	g_signal_connect_object (client, "map-info-event", 
	                         (GCallback) gsampler_midi_map_on_map_info, self, 0);
	g_signal_connect_object (client, "midi-instrument-info-event", 
	                         (GCallback) gsampler_midi_map_on_midi_instrument_info, self, 0);
	gtk_widget_set_sensitive ((GtkWidget*) self, TRUE);
	g_strfreev (map_list);
	g_object_unref (client);
}


void 
gsampler_midi_map_inactive (GsamplerMidiMap* self) {
	GsamplerClient* client = NULL;
	g_return_if_fail (self != NULL);
	
	client = gsampler_client_get_instance ();
	gsampler_midi_map_status = FALSE;
	g_signal_handlers_disconnect_by_func (client, 
	                                      gsampler_midi_map_on_map_count, self);
	g_signal_handlers_disconnect_by_func (client, 
	                                      gsampler_midi_map_on_midi_instrument_count, self);
	g_signal_handlers_disconnect_by_func (client, 
	                                      gsampler_midi_map_on_map_info, self);
	g_signal_handlers_disconnect_by_func (client, 
	                                      gsampler_midi_map_on_midi_instrument_info, self);
	gtk_widget_set_sensitive ((GtkWidget*) self, FALSE);
	g_object_unref (client);
}


gboolean 
gsampler_midi_map_delete_all_map (GsamplerMidiMap* self) {
	gboolean result = FALSE;
	GsamplerClient* client;
	g_return_val_if_fail (self != NULL, FALSE);
	
	client = gsampler_client_get_instance ();
	result = gsampler_client_delete_all_midi_instrument_map (client);
	g_object_unref (client);
	return result;
}